from .color import ColorOption, ColorParams, ColourOption, ColourParams
from .emoji import EmojiOption, EmojiParams
from .member import MemberOption, MemberParams

__all__ = (
    "ColorOption",
    "ColorParams",
    "ColourOption",
    "ColourParams",
    "EmojiOption",
    "EmojiParams",
    "MemberOption",
    "MemberParams",
)



from .concurrency_limiter import (
    CommandConcurrencyLimiter,
    ConcurrencyLimiter,
    channel_concurrency,
    custom_concurrency,
    global_concurrency,
    guild_concurrency,
    member_concurrency,
    user_concurrency,
)
from .hooks import (
    LimiterHook,
    bot_has_permissions,
    channel_limiter,
    custom_limiter,
    dm_only,
    global_limiter,
    guild_limiter,
    guild_only,
    has_permissions,
    member_limiter,
    owner_only,
    user_limiter,
)
from .loops import CronLoop, IntervalLoop, cron_loop, interval_loop
from .ratelimiter import RateLimiter, RateLimiterExhaustedError

__all__ = (
    "CommandConcurrencyLimiter",
    "ConcurrencyLimiter",
    "CronLoop",
    "IntervalLoop",
    "LimiterHook",
    "RateLimiter",
    "RateLimiterExhaustedError",
    "bot_has_permissions",
    "channel_concurrency",
    "channel_limiter",
    "cron_loop",
    "custom_concurrency",
    "custom_limiter",
    "dm_only",
    "global_concurrency",
    "global_limiter",
    "guild_concurrency",
    "guild_limiter",
    "guild_only",
    "has_permissions",
    "interval_loop",
    "member_concurrency",
    "member_limiter",
    "owner_only",
    "user_concurrency",
    "user_limiter",
)

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import arc

# Plugins are a way to organize commands and other functionality
plugin = arc.RESTPlugin("bar")


@plugin.include
@arc.slash_command("bar", "Bar command")
async def bar_cmd(ctx: arc.RESTContext) -> None:
    await ctx.respond("Bar!")


# This will be called when the extension is loaded
# A loader must be present for the extension to be valid
@arc.loader
def loader(client: arc.RESTClient) -> None:
    client.add_plugin(plugin)


# And this will be called when the extension is unloaded
# Including this is optional, but if not present, the extension cannot be unloaded
@arc.unloader
def unloader(client: arc.RESTClient) -> None:
    client.remove_plugin(plugin)



import hikari

import arc

# Error handling is vital to any application, and arc provides a simple way to handle errors
# Related documentation guide: https://arc.hypergonial.com/guides/error_handling

bot = hikari.GatewayBot("...")
client = arc.GatewayClient(bot)


@client.include
@arc.slash_command("name", "description")
async def error_command_func(
    ctx: arc.GatewayContext, recoverable: arc.Option[bool, arc.BoolParams("Is the error recoverable?")] = True
) -> None:
    if recoverable:
        raise RuntimeError("I'm an error!")
    else:
        raise Exception("I'm a fatal error!")


# You can add an error handler to a command, group, plugin, or the client
@error_command_func.set_error_handler
async def error_handler(ctx: arc.GatewayContext, exc: Exception) -> None:
    if isinstance(exc, RuntimeError):
        await ctx.respond("This is fine!")
        return
    # If you don't handle the error, you should re-raise it to send to the next error handler
    raise exc


# Setting the client error handler defines the global error handler
@client.set_error_handler
async def client_error_handler(ctx: arc.GatewayContext, exc: Exception) -> None:
    await ctx.respond("Oops, something went wrong!")

    # You should raise unhanded errors even in the global handler
    # so that a traceback is printed to the console by arc
    raise exc


# Errors are propagated from one error handler to the next until one handles it
# or you run out of error handlers

# Propagation order:
# Command -> Command Group -> Plugin -> Client

bot.run()



from __future__ import annotations

import logging
import typing as t

import hikari

from arc.abc.client import Client
from arc.context import Context
from arc.errors import NoResponseIssuedError
from arc.events import CommandErrorEvent, StartedEvent, StoppingEvent
from arc.internal.sigparse import parse_event_signature
from arc.internal.types import GatewayBotT, RESTBotT
from arc.plugin import GatewayPluginBase, RESTPluginBase

if t.TYPE_CHECKING:
    import alluka
    import typing_extensions as te

    from arc import AutodeferMode
    from arc.internal.types import EventCallbackT, EventT, ResponseBuilderT

__all__ = (
    "GatewayClient",
    "GatewayClientBase",
    "GatewayContext",
    "GatewayPlugin",
    "RESTClient",
    "RESTClientBase",
    "RESTContext",
    "RESTPlugin",
)


T = t.TypeVar("T")
P = t.ParamSpec("P")

logger = logging.getLogger(__name__)


class GatewayClientBase(Client[GatewayBotT]):
    """The base class for an arc client with Gateway support.
    This class primarily exists to allow for the creation of custom client types.

    For the default implementation of a Gateway client, see [`GatewayClient`][arc.client.GatewayClient].

    Parameters
    ----------
    app : hikari.GatewayBotAware
        The application this client is for.
    default_enabled_guilds : t.Sequence[hikari.Snowflakeish] | None
        The guilds that commands will be registered in by default
    autosync : bool
        Whether to automatically sync commands on startup
    autodefer : bool | AutodeferMode
        Whether to automatically defer responses
        This applies to all commands, and can be overridden on a per-command basis.
    default_permissions : hikari.Permissions | hikari.UndefinedType
        The default permissions for commands
        This applies to all commands, and can be overridden on a per-command basis.
    is_nsfw : bool
        Whether commands are NSFW
        This applies to all commands, and can be overridden on a per-command basis.
    integration_types : t.Sequence[hikari.ApplicationIntegrationType]
        The integration types that commands will support the installation of
        This applies to all commands, and can be overridden on a per-command basis.
    invocation_contexts : t.Sequence[hikari.ApplicationContextType]
        The context types that commands can be invoked in
        This applies to all commands, and can be overridden on a per-command basis
    provided_locales : t.Sequence[hikari.Locale] | None
        The locales that will be provided to the client by locale provider callbacks
    injector : alluka.Client | None
        If you already have an injector instance, you may pass it here.
        Otherwise, a new one will be created by default.
    """

    __slots__: t.Sequence[str] = ()

    def __init__(
        self,
        app: GatewayBotT,
        *,
        default_enabled_guilds: t.Sequence[hikari.Snowflakeish | hikari.PartialGuild]
        | hikari.UndefinedType = hikari.UNDEFINED,
        autosync: bool = True,
        autodefer: bool | AutodeferMode = True,
        default_permissions: hikari.Permissions | hikari.UndefinedType = hikari.UNDEFINED,
        is_nsfw: bool = False,
        integration_types: t.Sequence[hikari.ApplicationIntegrationType] = [
            hikari.ApplicationIntegrationType.GUILD_INSTALL
        ],
        invocation_contexts: t.Sequence[hikari.ApplicationContextType] = [
            hikari.ApplicationContextType.GUILD,
            hikari.ApplicationContextType.PRIVATE_CHANNEL,
            hikari.ApplicationContextType.BOT_DM,
        ],
        provided_locales: t.Sequence[hikari.Locale] | None = None,
        injector: alluka.abc.Client | None = None,
    ) -> None:
        super().__init__(
            app,
            default_enabled_guilds=default_enabled_guilds,
            autosync=autosync,
            autodefer=autodefer,
            default_permissions=default_permissions,
            is_nsfw=is_nsfw,
            integration_types=integration_types,
            invocation_contexts=invocation_contexts,
            provided_locales=provided_locales,
            injector=injector,
        )
        self.app.event_manager.subscribe(hikari.StartedEvent, self._on_gatewaybot_startup)
        self.app.event_manager.subscribe(hikari.StoppingEvent, self._on_gatewaybot_shutdown)
        self.app.event_manager.subscribe(hikari.InteractionCreateEvent, self._on_gatewaybot_interaction_create)

    @property
    def is_rest(self) -> bool:
        """Whether this client is a REST client."""
        return False

    @property
    def cache(self) -> hikari.api.Cache:
        """The cache for this client."""
        return self.app.cache

    async def _on_gatewaybot_startup(self, event: hikari.StartedEvent) -> None:
        await self._on_startup()
        self.app.event_manager.dispatch(StartedEvent(self), return_tasks=False)

    async def _on_gatewaybot_shutdown(self, event: hikari.StoppingEvent) -> None:
        await self._on_shutdown()
        self.app.event_manager.dispatch(StoppingEvent(self), return_tasks=False)

    async def _on_gatewaybot_interaction_create(self, event: hikari.InteractionCreateEvent) -> None:
        if isinstance(event.interaction, hikari.CommandInteraction):
            await self.on_command_interaction(event.interaction)
        elif isinstance(event.interaction, hikari.AutocompleteInteraction):
            await self.on_autocomplete_interaction(event.interaction)

    async def _on_error(self, ctx: Context[te.Self], exception: Exception) -> None:
        if not self.app.event_manager.get_listeners(CommandErrorEvent):
            return await super()._on_error(ctx, exception)

        self.app.event_manager.dispatch(CommandErrorEvent(self, ctx, exception), return_tasks=False)

    def subscribe(self, event_type: type[EventT], callback: EventCallbackT[EventT]) -> None:
        """Subscribe to an event.

        Parameters
        ----------
        event_type : type[EventT]
            The event type to subscribe to.

            `EventT` must be a subclass of `hikari.events.base_events.Event`.
        callback : t.Callable[EventT], t.Awaitable[None]]
            The callback to call when the event is dispatched.
        """
        self.app.event_manager.subscribe(event_type, callback)  # pyright: ignore reportGeneralTypeIssues

    def unsubscribe(self, event_type: type[EventT], callback: EventCallbackT[EventT]) -> None:
        """Unsubscribe from an event.

        Parameters
        ----------
        event_type : type[EventT]
            The event type to unsubscribe from.
        callback : t.Callable[[EventT], t.Awaitable[None]]
            The callback to unsubscribe.
        """
        self.app.event_manager.unsubscribe(event_type, callback)  # pyright: ignore reportGeneralTypeIssues

    def listen(self, *event_types: type[EventT]) -> t.Callable[[EventCallbackT[EventT]], EventCallbackT[EventT]]:
        """Generate a decorator to subscribe a callback to an event type.

        This is a second-order decorator.

        Parameters
        ----------
        *event_types : type[EventT]
            The event types to subscribe to. If not provided, the event type will be inferred
            instead from the type hints on the function signature.

            `EventT` must be a subclass of `hikari.events.base_events.Event`.

        Returns
        -------
        t.Callable[t.Callable[[EventT], t.Awaitable[None]]], t.Callable[[EventT], t.Awaitable[None]]]
            A decorator for a coroutine function that passes it to
            `EventManager.subscribe` before returning the function
            reference.
        """

        def decorator(func: EventCallbackT[EventT]) -> EventCallbackT[EventT]:
            types = event_types or parse_event_signature(func)

            for event_type in types:
                self.subscribe(event_type, func)

            return func

        return decorator


class RESTClientBase(Client[RESTBotT]):
    """The base class for an arc client with REST support.
    This class primarily exists to allow for the creation of custom client types.

    For the default implementation of a REST client, see [`RESTClient`][arc.client.RESTClient].

    !!! warning
        The client will take over the `hikari.CommandInteraction` and `hikari.AutocompleteInteraction` listeners of the passed bot.

    Parameters
    ----------
    app : hikari.GatewayBotAware
        The application this client is for.
    default_enabled_guilds : t.Sequence[hikari.Snowflakeish | hikari.PartialGuild] | None
        The guilds that commands will be registered in by default
    autosync : bool
        Whether to automatically sync commands on startup
    autodefer : bool | AutodeferMode
        Whether to automatically defer responses
        This applies to all commands, and can be overridden on a per-command basis.
    default_permissions : hikari.Permissions | hikari.UndefinedType
        The default permissions for commands
        This applies to all commands, and can be overridden on a per-command basis.
    is_nsfw : bool
        Whether commands are NSFW
        This applies to all commands, and can be overridden on a per-command basis.
    integration_types : t.Sequence[hikari.ApplicationIntegrationType]
        The integration types that commands will support the installation of
        This applies to all commands, and can be overridden on a per-command basis.
    invocation_contexts : t.Sequence[hikari.ApplicationContextType]
        The context types that commands can be invoked in
        This applies to all commands, and can be overridden on a per-command basis
    provided_locales : t.Sequence[hikari.Locale] | None
        The locales that will be provided to the client by locale provider callbacks
    injector : alluka.abc.Client | None
        If you already have an injector instance, you may pass it here.
        Otherwise, a new one will be created by default.
    """

    __slots__: t.Sequence[str] = ()

    def __init__(
        self,
        app: RESTBotT,
        *,
        default_enabled_guilds: t.Sequence[hikari.Snowflakeish | hikari.PartialGuild]
        | hikari.UndefinedType = hikari.UNDEFINED,
        autosync: bool = True,
        autodefer: bool | AutodeferMode = True,
        default_permissions: hikari.Permissions | hikari.UndefinedType = hikari.UNDEFINED,
        integration_types: t.Sequence[hikari.ApplicationIntegrationType] = [
            hikari.ApplicationIntegrationType.GUILD_INSTALL
        ],
        invocation_contexts: t.Sequence[hikari.ApplicationContextType] = [
            hikari.ApplicationContextType.GUILD,
            hikari.ApplicationContextType.PRIVATE_CHANNEL,
            hikari.ApplicationContextType.BOT_DM,
        ],
        is_nsfw: bool = False,
        provided_locales: t.Sequence[hikari.Locale] | None = None,
        injector: alluka.abc.Client | None = None,
    ) -> None:
        super().__init__(
            app,
            default_enabled_guilds=default_enabled_guilds,
            autosync=autosync,
            autodefer=autodefer,
            default_permissions=default_permissions,
            is_nsfw=is_nsfw,
            integration_types=integration_types,
            invocation_contexts=invocation_contexts,
            provided_locales=provided_locales,
            injector=injector,
        )
        self.app.add_startup_callback(self._on_restbot_startup)
        self.app.add_shutdown_callback(self._on_restbot_shutdown)
        self.app.interaction_server.set_listener(
            hikari.CommandInteraction, self._on_restbot_interaction_create, replace=True
        )
        self.app.interaction_server.set_listener(
            hikari.AutocompleteInteraction, self._on_restbot_autocomplete_interaction_create, replace=True
        )

    @property
    def is_rest(self) -> bool:
        """Whether this client is a REST client."""
        return True

    async def _on_restbot_startup(self, bot: hikari.RESTBotAware) -> None:
        await self._on_startup()
        await self.on_restbot_startup(bot)

    async def on_restbot_startup(self, bot: hikari.RESTBotAware) -> None:
        """A function that is called when the client is started.

        Parameters
        ----------
        bot : hikari.RESTBotAware
            The bot that was started.
        """

    async def _on_restbot_shutdown(self, bot: hikari.RESTBotAware) -> None:
        await self._on_shutdown()
        await self.on_restbot_shutdown(bot)

    async def on_restbot_shutdown(self, bot: hikari.RESTBotAware) -> None:
        """A function that is called when the client is shut down.

        Parameters
        ----------
        bot : hikari.RESTBotAware
            The bot that was shut down.
        """

    async def _on_restbot_interaction_create(self, interaction: hikari.CommandInteraction) -> ResponseBuilderT:
        builder = await self.on_command_interaction(interaction)
        if builder is None:
            raise NoResponseIssuedError(
                f"No response was issued to interaction for command: {interaction.command_name} ({interaction.command_type})."
            )
        return builder

    async def _on_restbot_autocomplete_interaction_create(
        self, interaction: hikari.AutocompleteInteraction
    ) -> hikari.api.InteractionAutocompleteBuilder:
        builder = await self.on_autocomplete_interaction(interaction)
        if builder is None:
            raise NoResponseIssuedError(
                f"No response was issued to autocomplete request for command: {interaction.command_name} ({interaction.command_type})."
            )
        return builder


# These two are classes for DI reasons, they should not ever contain any code.
class GatewayClient(GatewayClientBase[hikari.GatewayBotAware]):
    """The default gateway client implementation. Effectively an alias for [`arc.GatewayClientBase[hikari.GatewayBotAware]`][arc.client.GatewayClientBase].

    If you want to use a RESTBot, use [`RESTClient`][arc.client.RESTClient] instead.

    Parameters
    ----------
    app : hikari.GatewayBotAware
        The application this client is for.
    default_enabled_guilds : t.Sequence[hikari.Snowflakeish] | None
        The guilds that commands will be registered in by default
    autosync : bool
        Whether to automatically sync commands on startup
    autodefer : bool | AutodeferMode
        Whether to automatically defer responses
        This applies to all commands, and can be overridden on a per-command basis.
    default_permissions : hikari.Permissions | hikari.UndefinedType
        The default permissions for commands
        This applies to all commands, and can be overridden on a per-command basis.
    is_nsfw : bool
        Whether commands are NSFW
        This applies to all commands, and can be overridden on a per-command basis.
    integration_types : t.Sequence[hikari.ApplicationIntegrationType]
        The integration types that commands will support the installation of
        This applies to all commands, and can be overridden on a per-command basis.
    invocation_contexts : t.Sequence[hikari.ApplicationContextType]
        The context types that commands can be invoked in
        This applies to all commands, and can be overridden on a per-command basis
    provided_locales : t.Sequence[hikari.Locale] | None
        The locales that will be provided to the client by locale provider callbacks
    injector : alluka.Client | None
        If you already have an injector instance, you may pass it here.
        Otherwise, a new one will be created by default.

    Example
    --------
    ```py
    import hikari
    import arc

    bot = hikari.GatewayBot("TOKEN")
    client = arc.GatewayClient(bot)
    ```
    """

    __slots__: t.Sequence[str] = ()


class RESTClient(RESTClientBase[hikari.RESTBotAware]):
    """The default REST client implementation. Effectively an alias for [`arc.RESTClientBase[hikari.RESTBotAware]`][arc.client.RESTClientBase].

    If you want to use GatewayBot, use [`GatewayClient`][arc.client.GatewayClient] instead.

    Parameters
    ----------
    app : hikari.GatewayBotAware
        The application this client is for.
    default_enabled_guilds : t.Sequence[hikari.Snowflakeish | hikari.PartialGuild] | None
        The guilds that commands will be registered in by default
    autosync : bool
        Whether to automatically sync commands on startup
    autodefer : bool | AutodeferMode
        Whether to automatically defer responses
        This applies to all commands, and can be overridden on a per-command basis.
    default_permissions : hikari.Permissions | hikari.UndefinedType
        The default permissions for commands
        This applies to all commands, and can be overridden on a per-command basis.
    is_nsfw : bool
        Whether commands are NSFW
        This applies to all commands, and can be overridden on a per-command basis.
    integration_types : t.Sequence[hikari.ApplicationIntegrationType]
        The integration types that commands will support the installation of
        This applies to all commands, and can be overridden on a per-command basis.
    invocation_contexts : t.Sequence[hikari.ApplicationContextType]
        The context types that commands can be invoked in
        This applies to all commands, and can be overridden on a per-command basis
    provided_locales : t.Sequence[hikari.Locale] | None
        The locales that will be provided to the client by locale provider callbacks
    injector : alluka.abc.Client | None
        If you already have an injector instance, you may pass it here.
        Otherwise, a new one will be created by default.


    Example
    --------
    ```py
    import hikari
    import arc

    bot = hikari.RESTBot("TOKEN")
    client = arc.RESTClient(bot)
    ```

    !!! warning
        The client will take over the `hikari.CommandInteraction` and `hikari.AutocompleteInteraction` listeners of the passed bot.
    """

    __slots__: t.Sequence[str] = ()


GatewayContext = Context[GatewayClient]
"""A context using the default gateway client implementation. An alias for [`arc.Context[arc.GatewayClient]`][arc.context.base.Context]."""

RESTContext = Context[RESTClient]
"""A context using the default REST client implementation. An alias for [`arc.Context[arc.RESTClient]`][arc.context.base.Context]."""

RESTPlugin = RESTPluginBase[RESTClient]
"""A plugin using the default REST client implementation. An alias for [`arc.RESTPluginBase[arc.RESTClient]`][arc.plugin.RESTPluginBase]."""

GatewayPlugin = GatewayPluginBase[GatewayClient]
"""An alias for [`arc.GatewayPluginBase[arc.GatewayClient]`][arc.plugin.GatewayPluginBase]."""

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

# Paste your bot token from the developer portal here in place of '...':
bot = hikari.GatewayBot("...")

# Initialize arc with the bot:
client = arc.GatewayClient(bot)


@client.include  # Add command to client
@arc.slash_command("hi", "Say hi to someone!")  # Define command
async def hi_slash(
    # The context contains information about the command invocation
    ctx: arc.GatewayContext,
    # To add an option to a command, use the following syntax:
    user: arc.Option[hikari.User, arc.UserParams("The user to say hi to.")],
) -> None:
    await ctx.respond(f"Hey {user.mention}!")


# This must be on the last line, no code will run after this:
bot.run()



import hikari

import arc

# arc should be installed with hikari-arc[rest] for this example to work.

# Note: RESTBots also support bearer tokens, along with bot tokens.

# Paste your token from the developer portal here in place of '...':
bot = hikari.RESTBot("...")

# Initialize arc with the bot:
client = arc.RESTClient(bot)


@client.include  # Add command to client
@arc.slash_command("hi", "Say hi to someone!")  # Define command
async def hi_slash(
    # The context contains information about the command invocation
    ctx: arc.RESTContext,
    # To add an option to a command, use the following syntax:
    user: arc.Option[hikari.User, arc.UserParams("The user to say hi to.")],
) -> None:
    await ctx.respond(f"Hey {user.mention}!")


# This must be on the last line, no code will run after this:
bot.run()

# Note:
# You should set the interaction URL in the developer portal to point the URL of your server.
# You may need a domain name for this, as Discord requires SSL. Hikari starts the server on port 8080 by default.

# Tip:
# For local development, certain editors such as VS Code allow you to create a temporary URL to forward a port through:
# https://code.visualstudio.com/docs/editor/port-forwarding (Do not forget to set the port publicly accessible!)

# You can also use localhost.run:
# https://localhost.run/

# This should allow you to test your bot locally without having to deploy it to a server.



import typing as t

import alluka
import hikari

import arc
from arc.internal.types import ResponseBuilderT


class MockClient(arc.abc.Client[t.Any]):
    def __init__(
        self,
        app: t.Any,
        *,
        default_enabled_guilds: t.Sequence[hikari.Snowflakeish | hikari.PartialGuild]
        | hikari.UndefinedType = hikari.UNDEFINED,
        autosync: bool = True,
        autodefer: bool | arc.AutodeferMode = True,
        default_permissions: hikari.Permissions | hikari.UndefinedType = hikari.UNDEFINED,
        is_nsfw: bool = False,
        integration_types: t.Sequence[hikari.ApplicationIntegrationType] = [
            hikari.ApplicationIntegrationType.GUILD_INSTALL
        ],
        invocation_contexts: t.Sequence[hikari.ApplicationContextType] = [
            hikari.ApplicationContextType.BOT_DM,
            hikari.ApplicationContextType.GUILD,
            hikari.ApplicationContextType.PRIVATE_CHANNEL,
        ],
        provided_locales: t.Sequence[hikari.Locale] | None = None,
        injector: alluka.abc.Client | None = None,
    ) -> None:
        super().__init__(
            app,
            default_enabled_guilds=default_enabled_guilds,
            autosync=autosync,
            autodefer=autodefer,
            default_permissions=default_permissions,
            is_nsfw=is_nsfw,
            integration_types=integration_types,
            invocation_contexts=invocation_contexts,
            provided_locales=provided_locales,
            injector=injector,
        )
        self._owner_ids: list[hikari.Snowflake] = [hikari.Snowflake(123456789)]

    @property
    def is_rest(self) -> bool:
        return True  # Setting it to True so we get back builders we can verify

    async def push_inter(self, interaction: hikari.CommandInteraction) -> ResponseBuilderT:
        builder = await self.on_command_interaction(interaction)
        if builder is None:
            raise arc.NoResponseIssuedError(
                f"No response was issued to interaction for command: {interaction.command_name} ({interaction.command_type})."
            )
        return builder

    async def push_autocomplete_inter(
        self, interaction: hikari.AutocompleteInteraction
    ) -> hikari.api.InteractionAutocompleteBuilder:
        builder = await self.on_autocomplete_interaction(interaction)
        if builder is None:
            raise arc.NoResponseIssuedError(
                f"No response was issued to autocomplete request for command: {interaction.command_name} ({interaction.command_type})."
            )
        return builder


class MockPlugin(arc.PluginBase[MockClient]):
    @property
    def is_rest(self) -> bool:
        return True


MockContext = arc.Context[MockClient]



from __future__ import annotations

import abc
import functools
import inspect
import itertools
import typing as t

import hikari

from arc.abc.command import CallableCommandBase, _CommandSettings
from arc.abc.concurrency_limiting import ConcurrencyLimiterProto, HasConcurrencyLimiter
from arc.abc.error_handler import HasErrorHandler
from arc.abc.hookable import Hookable
from arc.command import MessageCommand, SlashCommand, SlashGroup, UserCommand
from arc.command.slash import SlashSubCommand, SlashSubGroup
from arc.context import AutodeferMode, Context
from arc.internal.types import BuilderT, ClientT, ErrorHandlerCallbackT, HookT, PostHookT, SlashCommandLike

if t.TYPE_CHECKING:
    from arc.abc.command import CommandBase

__all__ = ("PluginBase",)

P = t.ParamSpec("P")
T = t.TypeVar("T")


class PluginBase(HasErrorHandler[ClientT], Hookable[ClientT], HasConcurrencyLimiter[ClientT]):
    """An abstract base class for plugins.

    !!! note
        Parameters left as `hikari.UNDEFINED` will be inherited from the parent client.

    Parameters
    ----------
    name : str
        The name of this plugin. This must be unique across all plugins.
    default_enabled_guilds : t.Sequence[hikari.Snowflake] | hikari.UndefinedType
        The default guilds to enable commands in
    autodefer : bool | AutodeferMode
        If True, all commands in this plugin will automatically defer if it is taking longer than 2 seconds to respond.
        This can be overridden on a per-command basis.
    integration_types : t.Sequence[hikari.ApplicationIntegrationType] | hikari.UndefinedType
        The integration types that commands will support the installation of
        This can be overridden on a per-command basis.
    invocation_contexts : t.Sequence[hikari.ApplicationContextType] | hikari.UndefinedType
        The context types that commands can be invoked in
        This can be overridden on a per-command basis.
    default_permissions : hikari.Permissions | hikari.UndefinedType
        The default permissions for this plugin
        This can be overridden on a per-command basis.
    is_nsfw : bool | hikari.UndefinedType
        Whether this plugin is only usable in NSFW channels
    """

    __slots__: t.Sequence[str] = (
        "_client",
        "_cmd_settings",
        "_concurrency_limiter",
        "_default_enabled_guilds",
        "_error_handler",
        "_hooks",
        "_message_commands",
        "_name",
        "_post_hooks",
        "_slash_commands",
        "_user_commands",
    )

    def __init__(
        self,
        name: str,
        *,
        default_enabled_guilds: t.Sequence[hikari.Snowflakeish | hikari.PartialGuild]
        | hikari.UndefinedType = hikari.UNDEFINED,
        autodefer: bool | AutodeferMode | hikari.UndefinedType = hikari.UNDEFINED,
        integration_types: t.Sequence[hikari.ApplicationIntegrationType] | hikari.UndefinedType = hikari.UNDEFINED,
        invocation_contexts: t.Sequence[hikari.ApplicationContextType] | hikari.UndefinedType = hikari.UNDEFINED,
        default_permissions: hikari.Permissions | hikari.UndefinedType = hikari.UNDEFINED,
        is_nsfw: bool | hikari.UndefinedType = hikari.UNDEFINED,
    ) -> None:
        self._client: ClientT | None = None
        self._name = name
        self._default_enabled_guilds = (
            tuple(hikari.Snowflake(i) for i in default_enabled_guilds)
            if default_enabled_guilds is not hikari.UNDEFINED
            else hikari.UNDEFINED
        )

        self._cmd_settings = _CommandSettings(
            autodefer=AutodeferMode(autodefer) if isinstance(autodefer, bool) else autodefer,
            default_permissions=default_permissions,
            integration_types=integration_types,
            invocation_contexts=invocation_contexts,
            is_nsfw=is_nsfw,
        )

        self._slash_commands: dict[str, SlashCommandLike[ClientT]] = {}
        self._user_commands: dict[str, UserCommand[ClientT]] = {}
        self._message_commands: dict[str, MessageCommand[ClientT]] = {}
        self._error_handler: ErrorHandlerCallbackT[ClientT] | None = None
        self._hooks: list[HookT[ClientT]] = []
        self._post_hooks: list[PostHookT[ClientT]] = []
        self._concurrency_limiter: ConcurrencyLimiterProto[ClientT] | None = None

    @property
    def error_handler(self) -> ErrorHandlerCallbackT[ClientT] | None:
        """The error handler for this plugin."""
        return self._error_handler

    @error_handler.setter
    def error_handler(self, callback: ErrorHandlerCallbackT[ClientT] | None) -> None:
        """Set the error handler for this plugin."""
        self._error_handler = callback

    @property
    def concurrency_limiter(self) -> ConcurrencyLimiterProto[ClientT] | None:
        """The concurrency limiter for this plugin."""
        return self._concurrency_limiter

    @concurrency_limiter.setter
    def concurrency_limiter(self, limiter: ConcurrencyLimiterProto[ClientT] | None) -> None:
        """Set the concurrency limiter for this plugin."""
        self._concurrency_limiter = limiter

    @property
    def hooks(self) -> t.MutableSequence[HookT[ClientT]]:
        """The pre-execution hooks for this plugin."""
        return self._hooks

    @property
    def post_hooks(self) -> t.MutableSequence[PostHookT[ClientT]]:
        """The post-execution hooks for this plugin."""
        return self._post_hooks

    @property
    @abc.abstractmethod
    def is_rest(self) -> bool:
        """Whether or not this plugin is a REST plugin."""

    @property
    def name(self) -> str:
        """The name of this plugin."""
        return self._name

    @property
    def client(self) -> ClientT:
        """The client this plugin is included in."""
        if self._client is None:
            raise RuntimeError(
                f"Plugin '{self.name}' was not included in a client, '{type(self).__name__}.client' cannot be accessed until it is included in a client."
            )
        return self._client

    @property
    def default_enabled_guilds(self) -> t.Sequence[hikari.Snowflake] | hikari.UndefinedType:
        """The default guilds to enable commands in."""
        return self._default_enabled_guilds

    async def _handle_exception(self, ctx: Context[ClientT], exc: Exception) -> None:
        try:
            if self.error_handler is not None:
                await ctx._injection_ctx.call_with_async_di(self.error_handler, ctx, exc)
            else:
                raise exc
        except Exception as exc:
            await self.client._on_error(ctx, exc)

    def _resolve_settings(self) -> _CommandSettings:
        settings = self._client._cmd_settings if self._client is not None else _CommandSettings.default()
        return settings.apply(self._cmd_settings)

    def _resolve_hooks(self) -> list[HookT[ClientT]]:
        return self.client._hooks + self._hooks

    def _resolve_post_hooks(self) -> list[PostHookT[ClientT]]:
        return self.client._post_hooks + self._post_hooks

    def _resolve_concurrency_limiter(self) -> ConcurrencyLimiterProto[ClientT] | None:
        return self._concurrency_limiter or self.client._concurrency_limiter

    def _client_include_hook(self, client: ClientT) -> None:
        if client._plugins.get(self.name) is not None:
            raise RuntimeError(f"Plugin '{self.name}' is already included in client.")

        self._client = client
        self._client._plugins[self.name] = self

        for command in itertools.chain(
            self._slash_commands.values(), self._user_commands.values(), self._message_commands.values()
        ):
            command._client_include_hook(client)

    def _client_remove_hook(self) -> None:
        if self._client is None:
            raise RuntimeError(f"Plugin '{self.name}' is not included in a client.")

        for command in itertools.chain(
            self._slash_commands.values(), self._user_commands.values(), self._message_commands.values()
        ):
            self.client._remove_command(command)

        self._client._plugins.pop(self.name)
        self._client = None

    def _add_command(self, command: CommandBase[ClientT, t.Any]) -> None:
        if isinstance(command, (SlashCommand, SlashGroup)):
            self._slash_commands[command.name] = command
        elif isinstance(command, UserCommand):
            self._user_commands[command.name] = command
        elif isinstance(command, MessageCommand):
            self._message_commands[command.name] = command
        else:
            raise TypeError(f"Unknown command type '{type(command).__name__}'.")

    @t.overload
    def include(
        self,
    ) -> t.Callable[[CallableCommandBase[ClientT, BuilderT]], CallableCommandBase[ClientT, BuilderT]]: ...

    @t.overload
    def include(self, command: CallableCommandBase[ClientT, BuilderT]) -> CallableCommandBase[ClientT, BuilderT]: ...

    def include(
        self, command: CallableCommandBase[ClientT, BuilderT] | None = None
    ) -> (
        CallableCommandBase[ClientT, BuilderT]
        | t.Callable[[CallableCommandBase[ClientT, BuilderT]], CallableCommandBase[ClientT, BuilderT]]
    ):
        """Add a command to this plugin.

        !!! note
            This should be the **last** (topmost) decorator on a command.

        Parameters
        ----------
        command : arc.CommandBase[ClientT, BuilderT]
            The command to include in this plugin.

        Raises
        ------
        RuntimeError
            If the command is already included in this plugin.
        """

        def decorator(command: CallableCommandBase[ClientT, BuilderT]) -> CallableCommandBase[ClientT, BuilderT]:
            if command.plugin is not None:
                raise ValueError(f"Command '{command.name}' is already registered with plugin '{command.plugin.name}'.")

            command._plugin_include_hook(self)
            return command

        if command is not None:
            return decorator(command)

        return decorator

    def include_slash_group(
        self,
        name: str,
        description: str = "No description provided.",
        *,
        guilds: t.Sequence[hikari.Snowflakeish | hikari.PartialGuild] | hikari.UndefinedType = hikari.UNDEFINED,
        autodefer: bool | AutodeferMode | hikari.UndefinedType = hikari.UNDEFINED,
        integration_types: t.Sequence[hikari.ApplicationIntegrationType] | hikari.UndefinedType = hikari.UNDEFINED,
        invocation_contexts: t.Sequence[hikari.ApplicationContextType] | hikari.UndefinedType = hikari.UNDEFINED,
        is_nsfw: bool | hikari.UndefinedType = hikari.UNDEFINED,
        default_permissions: hikari.Permissions | hikari.UndefinedType = hikari.UNDEFINED,
        name_localizations: t.Mapping[hikari.Locale, str] | None = None,
        description_localizations: t.Mapping[hikari.Locale, str] | None = None,
    ) -> SlashGroup[ClientT]:
        """Add a new slash command group to this client.

        Parameters
        ----------
        name : str
            The name of the slash command group.
        description : str
            The description of the slash command group.
        guilds : t.Sequence[hikari.Snowflake] | hikari.UndefinedType
            The guilds to register the slash command group in
        autodefer : bool | AutodeferMode
            If True, all commands in this group will automatically defer if it is taking longer than 2 seconds to respond.
            This can be overridden on a per-subcommand basis.
        invocation_contexts : t.Sequence[hikari.ApplicationContextType]
            The context types to enable the slash command group in
        integration_types : t.Sequence[hikari.ApplicationIntegrationType]
            The integration types to enable the slash command group in
        default_permissions : hikari.Permissions | hikari.UndefinedType
            The default permissions for the slash command group
        name_localizations : dict[hikari.Locale, str]
            The name of the slash command group in different locales
        description_localizations : dict[hikari.Locale, str]
            The description of the slash command group in different locales
        is_nsfw : bool
            Whether the slash command group is only usable in NSFW channels

        Returns
        -------
        SlashGroup[te.Self]
            The slash command group that was created.

        !!! note
            Parameters left as `hikari.UNDEFINED` will be inherited from the parent plugin or client.

        Example
        --------
        ```py
        group = client.include_slash_group("Group", "A group of commands.")

        @group.include
        @arc.slash_subcommand(name="Command", description="A command.")
        async def cmd(ctx: arc.GatewayContext) -> None:
            await ctx.respond("Hello!")
        ```
        """
        guild_ids = tuple(hikari.Snowflake(i) for i in guilds) if guilds is not hikari.UNDEFINED else hikari.UNDEFINED

        group: SlashGroup[ClientT] = SlashGroup(
            name=name,
            description=description,
            guilds=guild_ids,
            autodefer=AutodeferMode(autodefer) if isinstance(autodefer, bool) else autodefer,
            invocation_contexts=invocation_contexts,
            integration_types=integration_types,
            default_permissions=default_permissions,
            name_localizations=name_localizations or {},
            description_localizations=description_localizations or {},
            is_nsfw=is_nsfw,
        )
        group._plugin_include_hook(self)
        return group

    @t.overload
    def inject_dependencies(self, func: t.Callable[P, T]) -> t.Callable[P, T]: ...

    @t.overload
    def inject_dependencies(self) -> t.Callable[[t.Callable[P, T]], t.Callable[P, T]]: ...

    def inject_dependencies(
        self, func: t.Callable[P, T] | None = None
    ) -> t.Callable[P, T] | t.Callable[[t.Callable[P, T]], t.Callable[P, T]]:
        """First order decorator to inject dependencies into the decorated function.

        !!! warning
            This makes functions uncallable if the plugin is not added to a client.

        !!! note
            Command callbacks are automatically injected with dependencies,
            thus this decorator is not needed for them.

        Example
        --------
        ```py
        class MyDependency:
            def __init__(self, value: str):
                self.value = value

        client = arc.GatewayClient(...)
        client.set_type_dependency(MyDependency, MyDependency("Hello!"))
        client.load_extension("foo")

        # In 'foo':

        plugin = arc.GatewayPlugin("My Plugin")

        @plugin.inject_dependencies
        def my_func(dep: MyDependency = arc.inject()) -> None:
            print(dep.value) # Prints "Hello!"

        @arc.loader
        def load(client: arc.GatewayClient) -> None:
            client.add_plugin(plugin)
        ```

        See Also
        --------
        - [`Client.set_type_dependency`][arc.client.Client.set_type_dependency]
            A method to set dependencies for the client.
        """

        def decorator(func: t.Callable[P, T]) -> t.Callable[P, T]:
            if inspect.iscoroutinefunction(func):

                @functools.wraps(func)
                async def decorator_async(*args: P.args, **kwargs: P.kwargs) -> T:
                    if self._client is None:
                        raise RuntimeError(
                            f"Cannot inject dependencies into '{func.__name__}' before plugin '{self.name}' is included in a client."
                        )
                    return await self._client.injector.call_with_async_di(func, *args, **kwargs)

                return decorator_async  # pyright: ignore reportGeneralTypeIssues
            else:

                @functools.wraps(func)
                def decorator(*args: P.args, **kwargs: P.kwargs) -> T:
                    if self._client is None:
                        raise RuntimeError(
                            f"Cannot inject dependencies into '{func.__name__}' before plugin '{self.name}' is included in a client."
                        )
                    return self._client.injector.call_with_di(func, *args, **kwargs)

                return decorator

        if func is not None:
            return decorator(func)

        return decorator

    @t.overload
    def find_command(
        self, command_type: t.Literal[hikari.CommandType.USER], full_name: str
    ) -> UserCommand[ClientT] | None: ...

    @t.overload
    def find_command(
        self, command_type: t.Literal[hikari.CommandType.MESSAGE], full_name: str
    ) -> MessageCommand[ClientT] | None: ...

    @t.overload
    def find_command(
        self, command_type: t.Literal[hikari.CommandType.SLASH], full_name: str
    ) -> SlashCommand[ClientT] | SlashSubCommand[ClientT] | SlashGroup[ClientT] | SlashSubGroup[ClientT] | None: ...

    def find_command(self, command_type: hikari.CommandType, full_name: str) -> t.Any | None:
        """Find a given command by it's fully qualified name.

        For instance, to locate a slash subcommand with the name `foo` in a group `bar`, you would pass `bar foo`.

        Parameters
        ----------
        command_type : hikari.CommandType
            The type of command to search for.
        full_name : str
            The fully qualified name of the command.

        Returns
        -------
        t.Any | None
            The command if found, otherwise None.
        """
        if command_type is hikari.CommandType.MESSAGE:
            return self._message_commands.get(full_name)
        if command_type is hikari.CommandType.USER:
            return self._user_commands.get(full_name)

        if command_type is not hikari.CommandType.SLASH:
            return None

        command_parts = full_name.split(" ")

        if len(command_parts) == 1:
            return self._slash_commands.get(command_parts[0])

        base_cmd = self._slash_commands.get(command_parts[0])

        if not isinstance(base_cmd, SlashGroup):
            return None

        subcmd = base_cmd.children.get(command_parts[1])

        if len(command_parts) == 2:
            return subcmd

        if not isinstance(subcmd, SlashSubGroup):
            return None

        return subcmd.children.get(command_parts[2])

    @t.overload
    def walk_commands(
        self, command_type: t.Literal[hikari.CommandType.USER], *, callable_only: bool = False
    ) -> t.Iterator[UserCommand[ClientT]]: ...

    @t.overload
    def walk_commands(
        self, command_type: t.Literal[hikari.CommandType.MESSAGE], *, callable_only: bool = False
    ) -> t.Iterator[MessageCommand[ClientT]]: ...

    @t.overload
    def walk_commands(
        self, command_type: t.Literal[hikari.CommandType.SLASH], *, callable_only: t.Literal[False]
    ) -> t.Iterator[
        SlashCommand[ClientT] | SlashSubCommand[ClientT] | SlashGroup[ClientT] | SlashSubGroup[ClientT]
    ]: ...

    @t.overload
    def walk_commands(
        self, command_type: t.Literal[hikari.CommandType.SLASH], *, callable_only: t.Literal[True]
    ) -> t.Iterator[SlashCommand[ClientT] | SlashSubCommand[ClientT]]: ...

    def walk_commands(  # noqa: C901
        self, command_type: hikari.CommandType, *, callable_only: bool = False
    ) -> t.Iterator[t.Any]:
        """Iterate over all commands of a certain type added to this plugin.

        Parameters
        ----------
        command_type : hikari.CommandType
            The type of commands to return.
        callable_only : bool
            Whether to only return commands that are directly callable.
            If True, command groups and subgroups will be skipped.

        Yields
        ------
        CommandT[ClientT]
            The next command that matches the given criteria.

        Example
        --------
        ```py
        for cmd in plugin.walk_commands(hikari.CommandType.SLASH):
            print(cmd.name)
        ```

        !!! tip
            To iterate over all types of commands, you may use [`itertools.chain()`][itertools.chain]:

            ```py
            import itertools

            for cmd in itertools.chain(
                plugin.walk_commands(hikari.CommandType.SLASH),
                plugin.walk_commands(hikari.CommandType.MESSAGE),
                plugin.walk_commands(hikari.CommandType.USER),
            ):
                print(cmd.name)
            ```
        """
        if hikari.CommandType.SLASH is command_type:
            for command in self._slash_commands.values():
                if isinstance(command, SlashCommand):
                    yield command
                    continue

                if not callable_only:
                    yield command

                for sub in command.children.values():
                    if isinstance(sub, SlashSubCommand):
                        yield sub
                        continue

                    if not callable_only:
                        yield sub

                    for subsub in sub.children.values():
                        yield subsub

        elif hikari.CommandType.MESSAGE is command_type:
            for command in self._message_commands.values():
                yield command

        elif hikari.CommandType.USER is command_type:
            for command in self._user_commands.values():
                yield command


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



from __future__ import annotations

import abc
import typing as t

from arc.internal.types import ClientT, HookableT, HookT, PostHookT

if t.TYPE_CHECKING:
    import typing_extensions as te


@t.final
class HookResult:
    """The result of a hook.

    Parameters
    ----------
    abort : bool
        Whether to abort the execution of the command.
        If True, the command execution will be silently aborted.
        If this is undesired, you should raise an exception instead.
    """

    __slots__: t.Sequence[str] = ("_abort",)

    def __init__(self, abort: bool = False) -> None:
        self._abort = abort


class Hookable(abc.ABC, t.Generic[ClientT]):
    """A trait for objects that can have hooks added to them."""

    __slots__: t.Sequence[str] = ()

    @property
    @abc.abstractmethod
    def hooks(self) -> t.MutableSequence[HookT[ClientT]]:
        """The pre-execution hooks for this object."""

    @property
    @abc.abstractmethod
    def post_hooks(self) -> t.MutableSequence[PostHookT[ClientT]]:
        """The post-execution hooks for this object."""

    def _resolve_hooks(self) -> list[HookT[ClientT]]:
        """Resolve all pre-execution hooks that apply to this object."""
        ...

    def _resolve_post_hooks(self) -> list[PostHookT[ClientT]]:
        """Resolve all post-execution hooks that apply to this object."""
        ...

    def add_hook(self, hook: HookT[ClientT]) -> te.Self:
        """Add a new pre-execution hook to this object.

        Any function that takes a [`Context`][arc.context.base.Context] as its sole parameter
        and returns either a [`HookResult`][arc.abc.hookable.HookResult] or
        `None` can be used as a hook.

        Parameters
        ----------
        hook : HookT[ClientT]
            The hook to add.

        Returns
        -------
        te.Self
            This object for chaining.
        """
        self.hooks.append(hook)
        return self

    def add_post_hook(self, hook: PostHookT[ClientT]) -> te.Self:
        """Add a new post-execution hook to this object.

        Any function that takes a [`Context`][arc.context.base.Context] as its sole parameter
        and returns `None` can be used as a post-hook.

        Parameters
        ----------
        hook : PostHookT[ClientT]
            The post-execution hook to add.

        Returns
        -------
        te.Self
            This object for chaining.
        """
        self.post_hooks.append(hook)
        return self


def with_hook(hook: HookT[ClientT]) -> t.Callable[[HookableT], HookableT]:
    """Add a new pre-execution hook to a hookable object. It will run before the command callback.

    Any function that takes a [`Context`][arc.context.base.Context] as its sole parameter
    and returns either a [`HookResult`][arc.abc.hookable.HookResult] or
    `None` can be used as a hook.

    Example
    --------
    ```py
    @client.include
    @arc.with_hook(arc.guild_only) # Add a pre-execution hook to a command
    @arc.slash_command("foo", "Foo command description")
    async def foo(ctx: arc.GatewayContext) -> None:
        ...
    ```
    """

    def decorator(hookable: HookableT) -> HookableT:
        hookable.hooks.append(hook)
        return hookable

    return decorator


def with_post_hook(hook: PostHookT[ClientT]) -> t.Callable[[HookableT], HookableT]:
    """Add a new post-execution hook to a hookable object. It will run after the command callback.

    Any function that takes a [`Context`][arc.context.base.Context] as its sole parameter
    and returns `None` can be used as a post-hook.

    Post-execution hooks are not executed if a pre-execution hook aborts the execution of the command.

    !!! warning
        Post-execution hooks **are** called even if the command callback raises an exception.
        You can see if the command callback failed by checking [`Context.has_command_failed`][arc.context.base.Context.has_command_failed].

    Example
    --------
    ```py
    @client.include
    @arc.with_post_hook(arc.guild_only) # Add a post-execution hook to a command
    @arc.slash_command("foo", "Foo command description")
    async def foo(ctx: arc.GatewayContext) -> None:
        ...
    ```
    """

    def decorator(hookable: HookableT) -> HookableT:
        hookable.post_hooks.append(hook)
        return hookable

    return decorator


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import os
import typing as t

import nox
from nox import options

PATH_TO_PROJECT = os.path.join(".", "arc")
EXAMPLES_PATH = os.path.join(".", "examples")
SCRIPT_PATHS = [PATH_TO_PROJECT, EXAMPLES_PATH, "noxfile.py", os.path.join(".", "tests")]

options.default_venv_backend = "uv"
options.sessions = ["format_fix", "slotscheck", "pyright", "pytest", "docs"]


# uv_sync taken from: https://github.com/hikari-py/hikari/blob/master/pipelines/nox.py#L48
#
# Copyright (c) 2020 Nekokatt
# Copyright (c) 2021-present davfsa
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
def uv_sync(
    session: nox.Session, /, *, include_self: bool = False, extras: t.Sequence[str] = (), groups: t.Sequence[str] = ()
) -> None:
    if extras and not include_self:
        raise RuntimeError("When specifying extras, set `include_self=True`.")

    args: list[str] = []
    for extra in extras:
        args.extend(("--extra", extra))

    group_flag = "--group" if include_self else "--only-group"
    for group in groups:
        args.extend((group_flag, group))

    session.run_install(
        "uv", "sync", "--frozen", *args, silent=True, env={"UV_PROJECT_ENVIRONMENT": session.virtualenv.location}
    )


@nox.session()
def format_fix(session: nox.Session) -> None:
    uv_sync(session, groups=["dev"])
    session.run("python", "-m", "ruff", "format", *SCRIPT_PATHS)
    session.run("python", "-m", "ruff", "check", *SCRIPT_PATHS, "--fix")


@nox.session()
def format(session: nox.Session) -> None:
    uv_sync(session, groups=["dev"])
    session.run("python", "-m", "ruff", "format", *SCRIPT_PATHS, "--check")
    session.run("python", "-m", "ruff", "check", *SCRIPT_PATHS)


@nox.session()
def pyright(session: nox.Session) -> None:
    uv_sync(session, include_self=True, groups=["dev"], extras=["cron"])
    session.run("pyright", *SCRIPT_PATHS)


@nox.session()
def pytest(session: nox.Session) -> None:
    uv_sync(session, include_self=True, groups=["dev"], extras=["cron"])
    session.run("pytest", "tests")


@nox.session()
def slotscheck(session: nox.Session) -> None:
    uv_sync(session, include_self=True, groups=["dev"], extras=["cron"])
    session.run("slotscheck", "-m", "arc", "-v")


@nox.session()
def docs(session: nox.Session) -> None:
    uv_sync(session, include_self=True, groups=["docs"])
    session.run("python", "-m", "mkdocs", "-q", "build")


@nox.session()
def servedocs(session: nox.Session) -> None:
    uv_sync(session, include_self=True, groups=["docs"])
    session.run("python", "-m", "mkdocs", "serve")


# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.



import hikari

import arc

plugin = arc.GatewayPlugin("test_plugin")


@plugin.include
@arc.slash_command("foo", "Test description")
async def foo_command(
    ctx: arc.GatewayContext,
    a: arc.Option[float | None, arc.FloatParams(description="foo", max=50.0)],
    b: arc.Option[hikari.GuildChannel | None, arc.ChannelParams(description="bar")],
) -> None:
    pass


@plugin.include
@arc.slash_command("bar", "Test description")
async def bar_command(
    ctx: arc.GatewayContext,
    a: arc.Option[hikari.Role | hikari.User, arc.MentionableParams(description="foo")],
    b: arc.Option[hikari.Attachment | None, arc.AttachmentParams(description="bar")],
) -> None:
    pass


@plugin.include
@arc.slash_command("baz", "Test description")
async def baz_command(
    ctx: arc.GatewayContext,
    a: arc.Option[int, arc.IntParams(description="foo", min=10)],
    b: arc.Option[str, arc.StrParams(description="bar", min_length=100)],
) -> None:
    pass


group = plugin.include_slash_group(
    "my_group", "My group description", default_permissions=hikari.Permissions.ADMINISTRATOR
)


@group.include
@arc.slash_subcommand("test_subcommand", "My subcommand description")
async def my_subcommand(
    ctx: arc.GatewayContext,
    a: arc.Option[int, arc.IntParams(description="foo", min=10)],
    b: arc.Option[str, arc.StrParams(description="bar", min_length=100)],
) -> None:
    pass


@arc.loader
def loader(client: arc.GatewayClient) -> None:
    client.add_plugin(plugin)


@arc.unloader
def unloader(client: arc.GatewayClient) -> None:
    client.remove_plugin(plugin)



"""CLI entry point for arc."""

import os
import platform
import sys

import hikari

import arc

# Support color on Windows
if sys.platform == "win32":
    import colorama

    colorama.init()


TEAL = "\x1b[38;5;79m"
WHITE = "\x1b[37m"

uname = platform.uname()
system_details = f"{uname.system} {uname.machine} ({uname.node}) - {uname.release}"
python_details = f"{platform.python_implementation()} {platform.python_version()} ({platform.python_compiler()})"

sys.stderr.write(
    f"""{TEAL}hikari-arc - package information
{WHITE}--------------------------------
{TEAL}Arc version: {WHITE}{arc.__version__}
{TEAL}Install path: {WHITE}{os.path.abspath(os.path.dirname(__file__))}
{TEAL}Hikari version: {WHITE}{hikari.__version__}
{TEAL}Install path: {WHITE}{os.path.abspath(os.path.dirname(hikari.__file__))}
{TEAL}Python: {WHITE}{python_details}
{TEAL}System: {WHITE}{system_details}\n\n"""
)

# MIT License
#
# Copyright (c) 2023-present hypergonial
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
